// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: ConstraintsChecker.c,v 1.61.206.6 2014/12/12 19:42:33 jcorbett Exp $
//
// R#report_range_constraints, R#check_package_pins,
// R#report_io_assignments, R#check_io_banks - Read a user constraint
// file (UCF) and analyze the locations of area groups, package pins,
// and IO buffers.

#include "ConstraintsChecker.h"

#include "Portability/export/basutstr.h"
#include "Xdm/export/xdm.h"
#include "Utilities/export/UtilSharedObject.h"
#include "ConstraintSystem/export/Cs.h"
#include "XOS/Portability/export/Port_StringBase.h"
#include "ConstraintSystem/export/Cs_Constraint_AreaGroup.h"
#include "ConstraintSystem/export/Cs_Constraint_Range.h"
#include "ConstraintSystem/export/Cs_Constraint_Locate.h"
#include "ConstraintSystem/export/Cs/Cs_ResSpec_Utils.h"
#include "ConstraintSystem/export/Cs/Cs_ResSpec_TypeName.h"
#include "ConstraintSystem/export/Cs_DeviceHelpers.h"
#include "ConstraintSystem/export/Cs_DeviceHelpers/Cs_DeviceResolver.h"
#include "TUtilities/export/regex.hpp"
#include "Databases/Rf/export/Facade/Rf_DeviceMgr.h"
#include "Databases/Rf/export/Helpers/Rf_SiteHelpers.h"
#include "GenericAlgorithms/Utilities/export/UtilCoord.h"
#include "Portability/export/Port_MsgDispatcher.h"

#include <stdlib.h>
#include <string>
#include <cstring>
#include <algorithm>

#include "PinGrid.h"
#include "ArgParser.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.61.206.6 $";

// Load the device model, initialize the pin grid and load the user
// constraints file.
ConstraintsChecker::ConstraintsChecker(Data& cadData)
   : data(cadData),
     deviceModel(data.GetDeviceModel()),
     pinGrid(new PinGrid(deviceModel)),
     errorCount(0), constraintMgr(0), xdm(0), design(0)
{
   LoadUCF();
   CollectAreaRanges();
}

// Tokenize a UCF file so we can filter unsupported statements before
// they are seen by the Constraint System. This is necessary because
// the Early Access Partial Reconfiguration software uses an older
// parser for UCF files and the PR-related mods have not yet been
// ported to the newer parser (Cs).
//
// To separate UCF files into statements, we need to find semicolons
// that are not contained in comments or quoted strings. There are
// three varieties of comments and three varieties of quoted strings.
//
// Inline- and block-style comments are supported. Either # or // can
// be used to begin an inline comment. The delimiters /* and */
// surround block-style comments. Despite the apparent similarity to
// C, block-style comments in UCF files can be nested.
//
// Quoted strings can be delimited by double quotes ("), single quotes
// ('), and back quotes (`). Quotation marks can be included within
// strings by escaping them with with a backslash (\).
//
// This file can be reverted to an earlier version that uses
// LoadConstraints() once the PR software is ported to the new parser
// or the new parser is updated to support PR constraints.
//
// The parsing loop has three main states:
//
// State Top - not in a comment, not in a quote. On quotePat, enter
//   stateQuote. On commentInlinePat, consume up to end of line (append \n
//   to statement buffer). On commentOpenPat, enter State Comment.
//
// State Quote - in a quoted string. On escapePat, append match to
//   statement buffer. On quotePat, append the quote to the statement
//   buffer. If the quote matches the opening quote, enter State Top.
//
// State Comment - in block comment. On commentOpenPat, increment depth
//   (including the first time). On commentClosePat, decrement the
//   depth. If the depth is zero, append a space to the statement
//   buffer.
//
// Inline comments and escape sequences are handled immediately with
// no need for separate states.

const boost::regex quotePat("^[\"'`].*");          // matches ", ', or `
const boost::regex commentInlinePat("^(#|//).*");  // matches # or //
const boost::regex commentOpenPat("^/\\*.*");      // matches /*
const boost::regex commentClosePat("^\\*/.*");     // matches */
const boost::regex escapePat("^\\\\..*");          // for all x matches \x
const boost::regex separatorPat("^;.*");           // matches ;

class UcfStatementParser {
public:
   UcfStatementParser(ifstream& source)
      : source(source), cursor(0), getlineCounter(0), statementStartLine(1)
      {}
   pair<string, long> GetStatement() {
      string statement;
      char quoteDelimiter('\0');
      long commentDepth(0);
      char lastChar('\0');
      while (source && lastChar != ';') {
         string input;
         getline(source, input);
         ++getlineCounter;
         buffer += "\n" + input;
         string::iterator end(buffer.end());
         for (; cursor < buffer.size(); ++cursor) {
            string::iterator start(buffer.begin() + cursor);
            if (commentDepth == 0) {
               if (quoteDelimiter == '\0') {
                  // At top level. Seek quote, comment, or semicolon.
                  if (boost::regex_match(start, end, quotePat)) {
                     quoteDelimiter = *start;
                  } else if (boost::regex_match(start, end, commentInlinePat)){
                     size_t
                        commentLength(buffer.find('\n', cursor) - cursor + 1);
                     buffer.erase(cursor, commentLength);
                     // In case part of a statement immediately
                     // precedes the comment with no intervening
                     // whitespace, add a space to avoid concatenating tokens.
                     buffer += " ";
                  } else if (boost::regex_match(start, end, commentOpenPat)) {
                     statement += buffer.substr(0, cursor);
                     ++cursor;
                     commentDepth = 1;
                  } else if (boost::regex_match(start, end, separatorPat)) {
                     statement += buffer.substr(0, cursor + 1);
                     buffer.erase(0, cursor + 1);
                     cursor = 0;
                     break;
                  }
                  // Else no additional action needed to advance one character.
               } else {
                  // Inside a quoted string. Seek matching quote that is
                  // not escaped.
                  if (boost::regex_match(start, end, escapePat)) {
                     // Consume the escape character. Normal for loop
                     // increment will consume the escaped character.
                     ++cursor;
                  } else if (*start == quoteDelimiter) {
                     // Found close quote. Return to top level state.
                     quoteDelimiter = '\0';
                  }
               }
            } else {
               // Inside a block comment
               if (boost::regex_match(start, end, commentOpenPat)) {
                  ++commentDepth;
                  // Consume the slash. Normal for loop increment will
                  // consume the star.
                  ++cursor;
               } else if (boost::regex_match(start, end, commentClosePat)) {
                  --commentDepth;
                  if (commentDepth != 0) {
                     // Consume the star. Normal for loop increment will
                     // consume the slash.
                     ++cursor;
                  } else {
                     buffer.erase(0, cursor + 2);
                     // Block comments act as a separator. foo/**/bar
                     // is equivalent to foo bar, not foobar.
                     statement += " ";
                     cursor = 0;
                  }
               }
            }
         }

         if (!statement.empty()) {
            lastChar = statement.at(statement.size() - 1);
         }
      }
      if (!source && lastChar != ';') {
         statement += buffer;
      }
      long line(statementStartLine);
      statementStartLine = getlineCounter;
      return make_pair(statement, line);
   }
private:
   ifstream& source;
   string buffer;
   string::size_type cursor;
   long getlineCounter;
   long statementStartLine;
};

const string sp("\\s*");        // optional spaces
const string q("[\"'`]");       // a quote
const string any(".+");         // any non-empty string

// Match statements like
//   AREA_GROUP "foo" MODE=RECONFIG;
//   AREA_GROUP "foo" PLACE=EXPAND/NO;
//   AREA_GROUP "foo" ROUTING=EXPAND/NO;
//   AREA_GROUP "foo" ROUTING=NOBOUNDARYCROSS;
//   AREA_GROUP "foo" PLACE=RANGE/GT11_X0Y0-GT11_X0Y3;
//   AREA_GROUP "foo" ROUTING=CONTAINED;
//
//   AREA_GROUP "foo" PRIVATE=PLACE;
//   AREA_GROUP "foo" CONTAINED=ROUTE;
//   AREA_GROUP "foo" BOUNDARYCROSS=NO;
// This pattern may accept whitespace in places where
// the early access parser would not. IVT cannot check
// this kind of constraint in any case.

const string expandNo("EXPAND" + sp + "/" + sp + "NO");

const string
modeReconfig         ("(" "MODE"    + sp + "=" + sp + "RECONFIG" ")");
const string
placeExpandNo        ("(" "PLACE"   + sp + "=" + sp + expandNo + ")");
const string
routingExpandNo      ("(" "ROUTING" + sp + "=" + sp + expandNo + ")");
const string
placeRange           ("(" "PLACE"   + sp + "=" + sp + "RANGE" + sp
                      + "/" + any + ")");
const string
routingNoBoundaryCross("(" "ROUTING" + sp + "=" + sp + "NOBOUNDARYCROSS" ")");
const string
routingContained     ("(" "ROUTING" + sp + "=" + sp + "CONTAINED" ")");

// New patterns for XPartitions (ISE 11 and beyond)
const string
privatePlaced        ("(" "PRIVATE" + sp + "=" + sp + "PLACE" ")");
const string
containedRoute       ("(" "CONTAINED" + sp + "=" + sp + "ROUTE" ")");
const string
boundaryCrossNo      ("(" "BOUNDARYCROSS" + sp + "=" + sp + "NO" ")");

const string
ucfName              ("(('[^']+')|(\"[^\"]+\")|(`[^`]+`)|(\\s+[^'\"` ]+\\s+)");

const string
spq                  ("(\\s|`|'|\")");

const string
locRange             ("INST" + spq + any + spq + "LOC" + sp + "=" + any + ":"
                      + any);

const string
configProhibit       ("CONFIG" + sp + "PROHIBIT" + spq + any);

const string
offset               ("OFFSET" + any);

const string
configInternalVRef   ("CONFIG" + sp + "INTERNAL_VREF" + any);

const boost::regex
ucfFilterPat(sp
             + "("
                + "(" + "AREA_GROUP" + sp + q + any + q + sp
                      + "("
                      + modeReconfig
                      + "|"
                      + placeExpandNo
                      + "|"
                      + routingExpandNo
                      + "|"
                      + routingNoBoundaryCross
                      + "|"
                      + placeRange
                      + "|"
                      + routingContained
                      + "|"
                      + privatePlaced
                      + "|"
                      + containedRoute
                      + "|"
                      + boundaryCrossNo
                      + ")"
                      + "|"
                      + "(" + locRange + ")"
                + ")"
             + "|"
             + "(" + configProhibit + ")"
             + "|"
             + "(" + offset + ")"
             + "|"
             + "(" + configInternalVRef + ")"
             + ")"
             + sp + ";",
             boost::regex::normal | boost::regbase::icase);

// Load a UCF file.
//
// Turns out the Early Access Partial Reconfiguration software
// uses an older system to load constraints. Their extensions to the
// constraint file syntax have not been ported to the newer constraint
// system (Cs) yet. So, the code here passes the UCF information
// to the constraint manager object a statement at a time after
// filtering out unsupported constructs.
//
void ConstraintsChecker::LoadUCF() {
   // To use the Constraints System, we need an Xdm design. An empty
   // one will do.
   xdm = new Xdm_Model("UCF_Placeholder", "ucf", "1.1");
   xdm->CreateTxn("txn", Xdm_Txn::MODIFY);
   Xdm_DesignMgr* mgr = xdm->GetDesignMgr();
   design = mgr->CreateDesign("design");

   constraintMgr =
      Util::SharedObject<Cs_ConstraintMgr>::Key(design).Create(design);
   if (!constraintMgr) {
      FatalInternal(29, "Unable to obtain constraint manager.");
   }

   ifstream ucfSource;
   ucfSource.open(data.constraintsFile.c_str());
   if (!ucfSource) {
      Fatal(64, "Unable to open " + data.constraintsFile + ".\n"
            "Check the file path, and file permissions.");
   }

   UcfStatementParser parser(ucfSource);
   while (ucfSource) {
      pair<string,long> statementLine(parser.GetStatement());
      string statement(statementLine.first);
      long line(statementLine.second);
      if (!statement.empty()) {
         if (boost::regex_match(statement, ucfFilterPat)) {
            ;                   // Just ignore statement.
         } else {
            try {
               Cs_ReturnCode
                  rc(constraintMgr->CreateConstraints(
                        statement.c_str(),
                        Cs_ConstraintSyntax::CS_UCF));
               if (!rc.Successful()) {
                  // No constraints were loaded so we'll report all
                  // messages and return.
                  Report("Input Error:5: The file " + data.constraintsFile
                         + "\ndid not load successfully.");
                  for (Cs_ReturnCode::const_iterator statusIt(rc.begin());
                       statusIt != rc.end();
                       statusIt++)
                  {
                     Cs_Status* status(*statusIt);
                     if (status->GetSeverity() != Cs_Status::NORMAL) {
                        string
                           msg(UT_STR::LstrToMbstr(status->GetLocalizedMsg()));
                        Report("  Line " + ToString(line)
                               + ": " + msg);
                     }
                  }
                  throw false;
               }
            } catch (Cs_Exception const& e) {
               // Cs uses both return codes and exceptions. The
               // exceptions aren't documented.

               // A regular expression to match the non-whitespace
               // delimited middle of the string. Note the first and
               // last grouped patterns are not recorded in the match
               // results due to the "?:".
               boost::regex trimPat("(?:\\s|\n)*(.*\\S)(?:\\s|\n)*");

               // Trim whitespace around the statement.
               boost::smatch result;
               if (boost::regex_match(statement, result, trimPat)) {
                  statement = result.str(1);
               }
               string
                  debugMessage(e.DbgGetInfo() + string("\n") +
                               string(UT_STR::LstrToMbstr(e.getLocalized())));

               // Trim whitespace around the exception text.
               if (boost::regex_match(debugMessage, result, trimPat)) {
                  debugMessage = result.str(1);
               }
               if (e.IsFatal()) {
                  FatalInternal(85, "A fatal error has occurred in the UCF "
                                "parser:\n" + debugMessage + "\nLine: "
                                + ToString(line) + "\n" + statement);
               } else {
                  Warn(86, "UCF parser indicated a non-fatal "
                       "error condition:\n"
                       + debugMessage + "\nLine: " + ToString(line)
                       + "\n" + statement);
               }
            }
         }
      }
   }
}

ConstraintsChecker::~ConstraintsChecker() {
   if (design) {
      Util::SharedObject<Cs_ConstraintMgr>::
         Key(design).Release(constraintMgr);
   }
   delete xdm;
}

// R#report_io_assignments
const string delim("[|\\s]");  // a vertical bar or a space
// Typical value:
//   INST "ZeJS988[6]" LOC =   A3 | AREA_GROUP = "AG_group2" ;
// but could also be
//   INST "ZeJS988[6]" AREA_GROUP="AG_group2"|LOC=A3;
// Match results:
//   result.str(0) = "INST "ZeJS988[6]" LOC =   A3 | AREA_GROUP = "AG_group2" ;"
//   result.str(1) = "AG_group2"
const boost::regex
areaGroupLocPat(any + delim
                + sp + "AREA_GROUP" + sp + "=" + sp
                + q + "(.+)" + q + any,
                boost::regex::normal | boost::regbase::icase);

const string ConstraintsChecker::GetAreaGroupName(Cs_Constraint_Locate* loc) {
   // Unfortunately, the association between location constraints and
   // area groups is not available in the Constraint System. So, we'll
   // get it from the UCF source text. Note: this will not handle all
   // the cases, so we'll require the users to keep it simple and list
   // the area group on the line with each locate constraint and to
   // refrain from using wild-cards.
   //
   // A better approach here would be to use the Cs_ResSpec as a key to
   // join Locate constraints with Area_Group constraints.
   string areaGroupName("<missing area group>");
   const Cs_ConstraintSourceInfo& csi(loc->GetSourceInfo());
   const string source(csi.GetAttrib(Cs_ConstraintSourceInfo::SRC_TEXT));
   boost::smatch matchResult;
   if (boost::regex_match(source, matchResult, areaGroupLocPat)) {
      // The regex has 1 parenthesized expression and matchResult.str(0)
      // contains the entire match. An error here would indicate
      // either an incorrect regular expression or boost::regex bug.
      if (matchResult.size() != 2) {
         FatalInternal(21, "Area group not in expected form: "
                       + string((const char*) loc->GetMsgName()) + ".");
      }
      areaGroupName = matchResult.str(1);
   } else {
      // This is the common case. We don't need to get the isolation
      // group from the area group. Instead we get it directly from
      // the pin isolation group (PIG) file.
   }
   return areaGroupName;
}

const boost::regex
areaGroupRangePat(sp + "AREA_GROUP" + sp + q + "(.+)" + q + any,
                  boost::regex::normal | boost::regbase::icase);

// Extract the area group name from an area range constraint. If
// this function fails, it issues a warning and returns
// "<missing area group>".
const string ConstraintsChecker::GetAreaGroupName(Cs_Constraint_Range* range)
{
   string areaGroupName("<missing area group>");
   const Cs_ConstraintSourceInfo& csi(range->GetSourceInfo());
   const string source(csi.GetAttrib(Cs_ConstraintSourceInfo::SRC_TEXT));
   boost::smatch matchResult;
   if (boost::regex_match(source, matchResult, areaGroupRangePat)) {
      if (matchResult.size() != 2) {
         FatalInternal(32, "Area group not in expected form: "
                       + string((const char*) range->GetMsgName()) + ".");
      }
      areaGroupName = matchResult.str(1);
   } else {
      Warn(51, "Unable to determine area group for range "
           "constraint:\n"
           + string((const char*) range->GetMsgName())
           + ".\nContact Xilinx for assistance.");
   }
   return areaGroupName;
}

// For each range constraint, collect the coordinates, site type and
// group name in the data instance.  Note: the distance between area
// ranges is computed during reporting in ReportAreaFaults() below.
void ConstraintsChecker::CollectAreaRanges() {
   Cs_ConstraintMgr::Constraints
      cs(constraintMgr
         ->GetConstraints(Cs_Filter_ConstraintType(Cs_Constraint_Range::Type()),
                          Cs_ConstraintMgr::ANYSTATE));

   for (Cs_ConstraintMgr::Constraints::iterator csIt(cs.begin());
        csIt != cs.end();
        ++csIt)
   {
      Cs_Constraint_Range*
         range(dynamic_cast<Cs_Constraint_Range*>(*csIt));
      if (range) {
         // There may be multiple ranges separated by commas.
         const string ranges(range->GetValue() + string(","));
         string::size_type start(0);
         string::size_type end(ranges.find_first_of(","));

         while (start != end) {
            string rangeOrSite(ranges.substr(start, end - start));
            string areaGroupName(GetAreaGroupName(range));

            // Given a range like "SLICE_X0Y110:SLICE_X1Y127" and the
            // constraint it came from, parse the range string and
            // look up the names of the associated area group and
            // isolation group.
            DeviceModel::SiteRect siteRect;
            bool gotOne(deviceModel->ParseRangeString(siteRect,
                                                      areaGroupName,
                                                      rangeOrSite));
            if (gotOne) {
               data.AddRange(siteRect);
            } else {
               ;
               // Got something like FRAME_ECC. IVT will take of the
               // general routing. Device analysis and design analysis
               // will take care of this sort of special case logic
               // tile.
            }

            start = ranges.find_first_not_of(",", end);
            end = ranges.find_first_of(",", start);
         }
      } else {
         FatalInternal(31, "Unable to process range constraint.");
      }
   }
   data.DeduplicateRanges();
   // Verify that all isolation groups have at least one area range.
   for (Data::ConstraintMapIt cmIt(data.constraintMap.begin());
        cmIt != data.constraintMap.end();
        ++cmIt)
   {
      const string isolationGroup(cmIt->second);
      Data::SiteRectMapIt srmIt(data.siteRectMap.find(isolationGroup));
      if (srmIt == data.siteRectMap.end()) {
         FatalInput(151, "Isolation group: " + isolationGroup +
                    " has no area range constraints.\n"
                    "Compare the area group spelling on the command line"
                    " to the UCF.");
      }
   }
}

// R#check_package_pins, R#check_io_banks,
// R#report_io_assignments

// When we take isolation groups for pins from a separate file,
// several things could go wrong.

// The user defines a new pin in the UCF but not in the PIG. IVT
// should issue warnings for all LOC constraints that aren't specified
// in isolation groups.

// The user removes a pin from the UCF but not the PIG. There should
// be a warning for each. Somehow we have to compare the list of pins
// from the PIG to the list of pins from the UCF.

// There can be contradictions between the area group a pin is in in
// the UCF and the PIG.

// The user does not specify an isolation
// group for a pin in the PIG.
//
// Since an isolation group may not have any pins, it is not an error
// to omit the PIG file or omit an isolation group from the PIG.
// Should we require an empty section in the PIG file for isolation
// groups with no pins?

// Verify that any adjacent package pins or IO buffers are from the
// same isolation group.
//
// The location constraints give us the pin names.
// The device resolver will give us the cells corresponding to the pins.
void ConstraintsChecker::CheckPinLocations() {

   Section("Package Pins, I/O Buffers, and I/O Banks");

   Report("  Note: It is incumbent on the user to verify that pins "
          "connected\n"
          "  to ignored nets are correct. For example, pins may not be\n"
          "  directly connected to trusted bus macros, but pins may be\n"
          "  connected to clocks, power, and global resets.\n\n"
          "    Pin(col, row) Bank      I/O Buffer Isolation Group Net\n"
          "  --------------- ---- --------------- --------------- -------");

   bool reportedPin(false);
   Cs_ConstraintMgr::Constraints
      cs(constraintMgr->
         GetConstraints(Cs_Filter_ConstraintType(Cs_Constraint_Locate::Type()),
                        Cs_ConstraintMgr::ANYSTATE));
   for (Cs_ConstraintMgr::Constraints::iterator csIt(cs.begin());
        csIt != cs.end();
        ++csIt)
   {
      Cs_Constraint_Locate* loc(dynamic_cast<Cs_Constraint_Locate*>(*csIt));
      if (loc) {
         Cs_DeviceResolver::Sites
            cells(Cs_DeviceResolver::ResolveSites(loc->GetDeviceSpec(),
                                                  deviceModel->GetRfDevice(),
                                                  constraintMgr));
         if (cells.size() == 1) {
            // This is the net.
            const Cs_ResSpec_TypeName*
               tn(dynamic_cast<const Cs_ResSpec_TypeName*>
                  (&(loc->GetResSpec())));
            if (!tn) {          // R#report_errors
               // Bogus result from Constraint System?
               FatalInternal(19, "Unable to obtain Cs_ResSpec_TypeName.");
            }
            // Might return ERROR_0_1, ERROR_1_1, or ERROR_2_1. In other
            // contexts it could return * but I don't know if here, too.
            const string netName(tn->GetMatchName());
            const string
               packagePinName((const char*) loc->GetDeviceSpecAsString());
            const string areaGroup(GetAreaGroupName(loc));

            // What it would mean if there were more than one site? Wild-card?
            for (Cs_DeviceResolver::Sites::iterator cellsIt(cells.begin());
                 cellsIt != cells.end();
                 ++cellsIt)
            {
               Rf_Cell cell(*cellsIt);
               Rf_PackagePin pin(deviceModel->GetPackagePin(cell));
               if (pin) {
                  Data::PinGroupMapIt pgmIt(data.pinGroupMap.find(netName));
                  if (pgmIt != data.pinGroupMap.end()) {
                     string isolationGroup(pgmIt->second);
                     if (areaGroup != "<missing area group>") {
                        Data::ConstraintMap::iterator
                           cmIt(data.constraintMap.find(areaGroup));
                        if (cmIt != data.constraintMap.end()) {
                           string ucfIsolationGroup(cmIt->second);
                           if (ucfIsolationGroup != isolationGroup) {
                              Warn(45, "Isolation group "
                                   + isolationGroup
                                   + " declared in pin isolation group"
                                   " file\ndoes not match isolation group "
                                   + ucfIsolationGroup
                                   + " inferred from area group\n"
                                   "for pin " + packagePinName
                                   + " on net " + netName + ".");
                           }
                        } else {
                           // This might be expected if there are area
                           // groups used for other purposes.
                           Warn(52, "Unable to find isolation group "
                                "for area group " + areaGroup + ".");
                        }
                     } else {
                        // Trust the pin isolation group file.
                     }
                     // R#check_package_pins, R#check_io_banks
                     pinGrid->LabelPin(pin, netName, isolationGroup);
                     reportedPin = true;
                  } else {
                     if (!data.pinGroupMap.empty()) { // R#report_errors
                        Warn(47, "No isolation group specified for "
                             "pin " + packagePinName + " on net "
                             + netName + ".");
                     }
                  }
               } else {
                  // Not a problem. Just some other kind of location
                  // constraint, e.g. SLICE_X86Y94, GT11CLK_X0Y3,
                  // BUFGCTRL_X0Y29 or DCM_ADV_X0Y6, etc.
               }
            }
         } else if (cells.size() != 0) { // R#report_errors
            FatalInput(25, "More than one site was associated "
                       "with LOC constraint:\n"
                       + string((const char*) loc->GetMsgName())
                       + "\nCheck for duplicated LOC constraints, wild cards, "
                       "\nmissing delimiters, or other syntax errors.");
         }
      } else {                  // R#report_errors
         FatalInternal(30, "Unable to process LOC constraint.");
      }
   }
   if (!reportedPin) {
      Report("  No pins were found.");
   }
   // R#check_package_pins, R#check_io_banks, R#report_io_assignments
   pinGrid->CheckPinLocations();
}

void ConstraintsChecker::Check() {
   CheckPinLocations();
   pinGrid->CheckIoBanks();     // R#check_io_banks
   errorCount += pinGrid->GetErrorCount();
}

// R#check_package_pins, R#check_io_banks, R#report_range_constraints,
// R#report_io_assignments, R#report_errors
void ConstraintsChecker::ReportFaults(TileChecker& tileChecker) {
   if (data.pinGroupMap.empty()) {

      Summary("\n  I/O Isolation\n\n"
              "    No pin isolation groups were supplied.");

   } else {
      pinGrid->ReportFaults();
   }

   tileChecker.ReportAreaSeparation();
   tileChecker.ReportSpecialTileFaults();

   errorCount += tileChecker.GetErrorCount();
   Summary("\n  UCF Isolation Verification Summary\n\n"
           "    Total Isolation Violations: " + ToString(errorCount));

}

// Return 0 for success and 1 for failure.
int ConstraintsChecker::GetStatus() { return errorCount != 0; }


